# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import sys

from blue_krill.data_types.url import MutableURL


def remove_suffix(s: str, suffix: str) -> str:
    """If given string endswith `suffix`, remove the suffix and return the new string"""
    if sys.version_info >= (3, 9):
        return s.removesuffix(suffix)
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def remove_prefix(s: str, prefix: str) -> str:
    """If given string startswith `prefix`, remove the prefix and return the new string"""
    if sys.version_info >= (3, 9):
        return s.removeprefix(prefix)
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def desensitize_url(url: str) -> str:
    """Obscure the password field in the url

    >>> desensitize_url("scheme://username:password@hostname:6666/path/?foo=bar")
    "scheme://username:********@hostname:6666/path/?foo=bar"
    """
    return MutableURL(url).obscure()



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import dataclasses
from collections import OrderedDict
from enum import Enum as OrigEnum
from enum import EnumMeta, auto
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type


@dataclasses.dataclass(init=False)
class FeatureFlagField:
    label: str
    default: bool
    name: str

    def __init__(
        self,
        name: Optional[str] = None,
        label: Optional[str] = None,
        default: bool = False,
    ):
        """FeatureFlag 中的字段, 记录了 label、default 等属性

        :param label: 对这个 feature flag 的描述语句
        :param default: feature flag 的默认状态, 对于新引入的 feature flag, 该值建议为 False.
        :param name: 当前 Feature Flag 的名字, 当使用 register_ext_feature_flag 进行注册时, 必须提供该字段.
        """
        self.name = name or ""
        self.label = label or name or ""
        self.default = default

    def __set_name__(self, owner, name):
        """利用描述符协议, 往 FeatureFlagField 注入 FeatureFlag 的名字."""
        self.name = name
        if not self.label:
            self.label = self.name

    def __get__(self, instance, owner):
        """返回 FeatureFlag 的名称. 因为 FeatureFlag 的值就是他自身的名称."""
        return self.name

    def __str__(self):
        return self.name


class FeatureFlagMeta(type):
    _feature_flag_fields_: Dict[str, FeatureFlagField]

    def __new__(mcs, cls_name: str, bases, dct: Dict):
        _feature_flag_fields_ = {}
        for base in bases:
            _feature_flag_fields_.update(getattr(base, "_feature_flag_fields_", {}))

        for attr, field in dct.items():
            if not isinstance(field, FeatureFlagField):
                continue

            if _feature_flag_fields_.get(attr, field).label != field.label:
                raise ValueError("Two Feature Flags cannot be set to the same value.")
            if not attr:
                raise ValueError("Feature flag's name can not be empty.")

            _feature_flag_fields_[attr] = field

        dct["_feature_flag_fields_"] = _feature_flag_fields_
        return super().__new__(mcs, cls_name, bases, dct)

    def _get_feature_fields_(cls) -> Dict[str, FeatureFlagField]:
        return cls._feature_flag_fields_

    def __iter__(cls):
        for feature in cls._get_feature_fields_():
            yield feature


class FeatureFlag(str, metaclass=FeatureFlagMeta):
    def __new__(cls, value):
        """Cast a string into a predefined feature flag."""
        for field in cls._get_feature_fields_().values():
            if field.name == value:
                return value
        return cls._missing_(value)

    @classmethod
    def _missing_(cls, value) -> str:
        raise ValueError("%r is not a valid %s" % (value, cls.__name__))

    @classmethod
    def get_default_flags(cls) -> Dict[str, bool]:
        """Get the default user feature flags, client is safe to modify the result because it's a copy"""
        features = {field.name: field.default for field in cls._get_feature_fields_().values()}
        return features.copy()

    @classmethod
    def get_django_choices(cls) -> List[Tuple[str, str]]:
        """Get Django-Like Choices for this FeatureFlag Collection."""
        return [(field.name, field.label) for field in cls._get_feature_fields_().values()]

    @classmethod
    def get_feature_label(cls, feature: str) -> str:
        """Get the label of provided feature flag"""
        return cls._get_feature_fields_()[cls(feature)].label

    @classmethod
    def register_feature_flag(cls, field: FeatureFlagField):
        """注册额外的FeatureFlagField"""
        name = field.name
        if cls._feature_flag_fields_.get(name, field).label != field.label:
            raise ValueError("Two Feature Flags cannot be set to the same value.")
        if not name:
            raise ValueError("Feature flag's name can not be empty.")

        cls._feature_flag_fields_[name] = field
        setattr(cls, name, field)


# Treat `EnumField` as Any to make it pass some strict type checking, for example:
# when `foo = EnumField(...)` and `foo` was read via `SomeClass.foo.value`.
#
# TODO: Write a mypy plugin to provide better typing experience.
if TYPE_CHECKING:
    EnumFieldBase = Any
else:
    EnumFieldBase = object


class EnumField(EnumFieldBase):
    """Use it with `StructuredEnum` type

    :param real_value: the real value of enum member
    :param label: the label text of current enum value
    :param is_reserved: if current member was reserved, it will not be included in choices
    """

    def __init__(self, real_value: Any, label: Optional[str] = None, is_reserved: bool = False):
        self.real_value = real_value
        self.label = label
        self.is_reserved = is_reserved

    def set_label_if_empty(self, key: str):
        """Set field's label if not provided"""
        if not self.label:
            self.label = key.lower().replace("_", " ").capitalize()


class StructuredEnumMeta(EnumMeta):
    """The metaclass of StructuredEnum"""

    __field_members__: Dict[Type, Dict]

    def __new__(metacls, cls, bases, classdict):
        field_members = metacls.process_enum_fields(classdict)
        classdict["__field_members__"] = field_members
        return super().__new__(metacls, cls, bases, classdict)

    @staticmethod
    def process_enum_fields(classdict) -> Dict:
        """Iterate all enum members, transform them into `EnumField` objects and return as a dict.

        `EnumField` members in `classdict` will be replaced with their `real_value` attribute so `EnumMeta` can
        continue the initialization.
        """
        fields = OrderedDict()
        # Find out all `EnumField` instance, store them into class so we can use them later
        for key, member in classdict.items():
            # Ignore all private members
            if key.startswith("_"):
                continue

            # Turn regular enum member into EnumField instance
            if not isinstance(member, EnumField) and isinstance(member, (int, str, auto)):
                member = EnumField(member)
            if not isinstance(member, EnumField) or member.is_reserved:
                continue

            member.set_label_if_empty(key)
            fields[key] = member
            # Use dict's setitem method because setting value with `classdict[key]` is forbidden
            dict.__setitem__(classdict, key, member.real_value)
        return fields

    def get_field_members(cls) -> Dict:
        return cls.__field_members__


class StructuredEnum(OrigEnum, metaclass=StructuredEnumMeta):
    """Structured Enum type, providing extra features such as getting enum members as choices tuple"""

    @classmethod
    def get_django_choices(cls) -> List[Tuple[Any, str]]:
        """Get Django-Like Choices for all field members."""
        return cls.get_choices()

    @classmethod
    def get_choice_label(cls, value: Any) -> str:
        """Get the label of field member by value"""
        if isinstance(value, cls):
            value = value.value

        members = cls.get_field_members()
        for field in members.values():
            if value == field.real_value:
                return field.label

        return value

    @classmethod
    def get_labels(cls) -> List[str]:
        """Get the label list for all field members."""
        return [item[1] for item in cls.get_choices()]

    @classmethod
    def get_values(cls) -> List[Any]:
        """Get the value list for all field members."""
        return [item[0] for item in cls.get_choices()]

    @classmethod
    def get_choices(cls) -> List[Tuple[Any, str]]:
        """Get Choices for all field members."""
        members = cls.get_field_members()
        return [(field.real_value, field.label) for field in members.values()]



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from typing import Optional

from cryptography.fernet import Fernet

from blue_krill.encoding import force_bytes, force_text

logger = logging.getLogger(__name__)


def get_default_secret_key():
    try:
        from django.conf import settings

        return settings.BKKRILL_ENCRYPT_SECRET_KEY
    except ImportError:
        return None


def get_default_encrypt_cipher_type():
    """获取默认的加密算法"""
    try:
        from django.conf import settings

        return getattr(settings, 'ENCRYPT_CIPHER_TYPE', 'FernetCipher')
    except ImportError:
        return 'FernetCipher'


def encrypt_string(s: str, key: Optional[bytes] = None) -> str:
    """Shortcut function: encrypt a string into a fernet token"""
    key = key or get_default_secret_key()
    f = Fernet(key)
    value = f.encrypt(force_bytes(s))
    return force_text(value)


def decrypt_string(s: str, key: Optional[bytes] = None) -> str:
    """Shortcut function: decrypt a fernet token into a string"""
    key = key or get_default_secret_key()
    f = Fernet(key)
    value = f.decrypt(force_bytes(s))
    return force_text(value)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import base64
from typing import Union

from cryptography.hazmat.primitives.ciphers import Cipher

from blue_krill.encoding import force_bytes, force_text


def legacy_decrypt(text: Union[bytes, str], legacy_secret_key: Union[bytes, str]) -> str:
    """解密数据（仅为兼容旧版本，请不要使用）

    :param text: 待解密数据
    :param legacy_secret_key: 用于解密的 secret key
    """
    raw_data = base64.b64decode(text)
    decryptor = _get_cipher(force_bytes(legacy_secret_key)).decryptor()
    result = decryptor.update(raw_data) + decryptor.finalize()
    return force_text(result).strip()


def legacy_encrypt(text: Union[bytes, str], legacy_secret_key: Union[bytes, str]) -> str:
    """加密数据（仅为兼容旧版本，请不要使用）

    :param text: 待加密数据
    :param legacy_secret_key: 用于加密的 secret key
    """
    # Append blank space to extend text in order to make it's length match DES3 block-size(8)
    bytes_text = force_bytes(text)
    if len(bytes_text) % 8 != 0:
        bytes_text += (8 - len(bytes_text) % 8) * b' '

    encryptor = _get_cipher(force_bytes(legacy_secret_key)).encryptor()
    result = encryptor.update(bytes_text) + encryptor.finalize()
    return force_text(base64.b64encode(result))


def _get_cipher(key: bytes) -> Cipher:
    """获取 DES3 Cipher 对象"""
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

    algorithm = algorithms.TripleDES(key)
    cipher = Cipher(algorithm, modes.CBC(key[:8]), backend=default_backend())
    return cipher



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from typing import ClassVar, Dict, Optional

from bkcrypto import constants
from bkcrypto.contrib.django.ciphers import get_symmetric_cipher
from bkcrypto.symmetric.options import SM4SymmetricOptions
from cryptography.fernet import Fernet

from blue_krill.encoding import force_bytes, force_text
from blue_krill.encrypt.utils import get_default_encrypt_cipher_type, get_default_secret_key


class _Header:
    def __init__(self, header: str):
        self.header = header

    def add_header(self, text: str):
        return self.header + text

    def strip_header(self, text: str):
        # 兼容无 header 加密串
        if not self.contain_header(text):
            return text
        return text[len(self.header):]

    def contain_header(self, text: str) -> bool:
        return text.startswith(self.header)


class EncryptHandler:
    cipher_classes: ClassVar[Dict] = {}

    def __init__(self, encrypt_cipher_type: Optional[str] = None, secret_key: Optional[bytes] = None):
        self._encrypt_cipher_type = encrypt_cipher_type
        self.secret_key = secret_key or get_default_secret_key()

    @property
    def encrypt_cipher_type(self):
        return self._encrypt_cipher_type or get_default_encrypt_cipher_type()

    def encrypt(self, text: str) -> str:
        """根据指定加密算法，加密字段"""
        # 已加密则不处理
        for _, cls in self.cipher_classes.items():
            if cls.header.contain_header(text):
                return text

        # 根据加密类型配置选择不同的加密算法
        try:
            cipher_class = self.cipher_classes[self.encrypt_cipher_type]
        except KeyError:
            raise ValueError(f"Invalid cipher type: {self.encrypt_cipher_type}")
        else:
            cipher = cipher_class(self.secret_key)
            return cipher.encrypt(text)

    def decrypt(self, encrypted: str) -> str:
        """根据 header 解密"""
        for _, cls in self.cipher_classes.items():
            if cls.header.contain_header(encrypted):
                cipher = cls(self.secret_key)
                return cipher.decrypt(encrypted)
        # 若不包含头则直接返回
        return encrypted


def register_cipher(cls):
    EncryptHandler.cipher_classes[cls.__name__] = cls


@register_cipher
class FernetCipher:
    header = _Header("bkcrypt$")

    def __init__(self, secret_key: Optional[bytes] = None):
        self.secret_key = secret_key or get_default_secret_key()
        self.cipher = Fernet(self.secret_key)

    def encrypt(self, text: str) -> str:
        b_text = force_bytes(text)
        return self.header.add_header(force_text(self.cipher.encrypt(b_text)))

    def decrypt(self, encrypted: str) -> str:
        encrypted = self.header.strip_header(encrypted)

        b_encrypted = force_bytes(encrypted)
        return force_text(self.cipher.decrypt(b_encrypted))


@register_cipher
class SM4CTR:
    header = _Header("sm4ctr$")

    def __init__(self, secret_key: Optional[bytes] = None):
        self.secret_key = secret_key or get_default_secret_key()
        self.cipher = get_symmetric_cipher(
            common={"key": self.secret_key},
            cipher_type=constants.SymmetricCipherType.SM4.value,
            cipher_options={
                constants.SymmetricCipherType.SM4.value: SM4SymmetricOptions(mode=constants.SymmetricMode.CTR)
            },
        )

    def encrypt(self, text: str) -> str:
        return self.header.add_header((self.cipher.encrypt(text)))

    def decrypt(self, encrypted: str) -> str:
        encrypted = self.header.strip_header(encrypted)

        return self.cipher.decrypt(encrypted)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""

import base64
import sys

import click
import editor
from cryptography.fernet import InvalidToken

from blue_krill.encoding import force_bytes, force_text
from blue_krill.encrypt.utils import decrypt_string, encrypt_string

"""The command line tool for blue_krill.secure module, it's can do follow things:

- encrypt a plain text
- decrypt a encrypted token
- edit a encrypted fernet token and output the new result
"""
COMMAND_HELP = '''The command-line tool for blue_krill.secure module, supported env vars:

- BK_FERNET_KEY: the default fernet key for encryption/decryption, if not set, will prompt for user input
- EDITOR: the default editor for `bk-secure edit` command, recommended value: "vi", "vim"
'''


def validate_key(ctx, param, value) -> bytes:
    """Validate fernet key"""
    b_value = force_bytes(value)

    help_text = '"key" should be a base64 encoded 32-bit string'
    try:
        decoded_value = base64.b64decode(b_value)
    except Exception:
        raise click.BadParameter(help_text)
    if not len(decoded_value) == 32:
        raise click.BadParameter(help_text)

    return b_value


@click.group(help=COMMAND_HELP)
def main():
    """The main entrance"""
    pass


@main.command(help='Encrypt a string into fernet token')
@click.option('--key', envvar='BK_FERNET_KEY', prompt='Input fernet key', callback=validate_key)
@click.option('--input-string', prompt='Input string')
def encrypt(key, input_string):
    """Encrypt a given content"""
    target = encrypt_string(input_string.strip(), key=key)
    click.echo('The encrypted token is: ', nl=False)
    click.echo(click.style(target, fg='red'))


@main.command(help='Decrypt a fernet token into plaintext')
@click.option('--key', envvar='BK_FERNET_KEY', prompt='Input fernet key', callback=validate_key)
@click.option('--input-token', prompt='Input token')
def decrypt(key, input_token):
    """Decrypt a given fernet token"""
    try:
        target = decrypt_string(input_token.strip(), key=key)
    except InvalidToken:
        click.echo('Error: Token is invalid')
        sys.exit(1)

    click.echo('The decrypted result is: ', nl=False)
    click.echo(click.style(target, fg='blue'))


@main.command(help='Edit a fernet token')
@click.option('--key', envvar='BK_FERNET_KEY', prompt='Input fernet key', callback=validate_key)
@click.option('--input-token', prompt='Input token')
def edit(key, input_token):
    """Edit a given fernet token, will call external editor"""
    try:
        target = decrypt_string(input_token.strip(), key=key)
    except InvalidToken:
        click.echo('Error: Token is invalid')
        sys.exit(1)

    # Call external editor for editting value
    new_value = force_text(editor.edit(contents=target)).strip()
    encrypted_new_value = encrypt_string(new_value, key=key)

    # Display the plain new value
    click.echo('The new value is: ', nl=False)
    click.echo(click.style(new_value, fg='blue'))

    # Display the encrypted new value
    click.echo('The encrypted new value is: ', nl=False)
    click.echo(click.style(encrypted_new_value, fg='red'))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from typing import Any, Dict, List, Optional, Union

from django.utils.functional import cached_property

from blue_krill.data_types.url import MutableURL

try:
    import redis
    from redis import sentinel  # noqa
except ImportError as _e:
    raise ImportError('Error loading redis module: %s.\n' 'Did you install a suitable version for redis?' % _e) from _e

SentinelHost = Dict[str, Optional[Union[str, int]]]


class SentinelBackend:
    """Redis Sentinel Backend"""

    def __init__(self, url: str, master_name: str, sentinel_kwargs: Dict[str, Any], **connection_kwargs):
        """
        :param url: 格式如 sentinel://0.0.0.0:26347/3;sentinel://0.0.0.0:26347/3
        :param master_name: sentinel service name
        :param sentinel_kwargs: sentinel 相关配置, 如 password, master_name 等
        :param connection_kwargs: redis 连接设置
        """
        self.master_name = master_name
        self.sentinel_kwargs = sentinel_kwargs
        self.connection_kwargs = connection_kwargs or {}
        self._parse_from_url(url)

    @property
    def hosts(self) -> List[SentinelHost]:
        return self._hosts

    @property
    def password(self) -> Optional[str]:
        """redis 实例密码, 非 sentinel 密码"""
        return self._password

    @property
    def db(self) -> int:
        return self._db

    @cached_property
    def client(self) -> redis.Redis:
        self.connection_kwargs['db'] = self.db
        if self.password:
            self.connection_kwargs['password'] = self.password

        sentinel_instance = sentinel.Sentinel(
            [(cp['host'], cp['port']) for cp in self.hosts],
            sentinel_kwargs=self.sentinel_kwargs,
            **self.connection_kwargs,
        )
        return sentinel_instance.master_for(service_name=self.master_name)

    def _parse_from_url(self, url: str):
        """从 url 解析出主机端口等信息"""
        chunks = url.split(';')

        parts = MutableURL(url=chunks[0])
        self._validate_scheme(parts.scheme)
        hosts = [{'host': parts.hostname, 'port': parts.port}]

        self._db = self._extract_db(parts.path)
        self._password = parts.password

        for chunk in chunks[1:]:
            parts = MutableURL(url=chunk)
            self._validate_scheme(parts.scheme)
            hosts.append({'host': parts.hostname, 'port': parts.port})

        self._hosts = hosts

    def _extract_db(self, path: Optional[str]) -> int:
        """从 url path 中提取 redis db.

        :param path: url path
        :return: redis db num. 如果从 path 中无法解析出合法的 db num, 直接返回0
        """
        if not path:
            return 0

        if path.startswith('/'):
            path = path[1:]

        try:
            return int(path)
        except (ValueError, TypeError):
            return 0

    def _validate_scheme(self, scheme: str):
        if scheme != 'sentinel':
            raise ValueError('url must be sentinel scheme')



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from typing import Any, Dict, Optional, Union
from urllib.parse import SplitResult, parse_qsl, unquote, urlsplit


class MutableURL:
    _url: str
    _components: SplitResult
    _query: dict

    def __init__(self, url: Union[str, 'MutableURL']):
        if isinstance(url, MutableURL):
            self._url = url._url
        elif isinstance(url, str):
            self._url = url
        else:
            raise TypeError(f"Invalid type for MutableURL. Expected str or MutableURL, got {type(url)}")

    @property
    def components(self) -> SplitResult:
        if not hasattr(self, "_components"):
            self._components = urlsplit(self._url)
        return self._components

    @property
    def scheme(self):
        return self.components.scheme

    @property
    def username(self) -> Optional[str]:
        return self.components.username

    @property
    def password(self) -> Optional[str]:
        # Auto unescape password
        return unquote(self.components.password) if self.components.password else None

    @property
    def hostname(self) -> Optional[str]:
        return self.components.hostname

    @property
    def port(self) -> Optional[int]:
        return self.components.port

    @property
    def netloc(self) -> Optional[str]:
        return self.components.netloc

    @property
    def path(self) -> Optional[str]:
        return unquote(self.components.path) if self.components.path else None

    @property
    def query(self) -> Dict:
        if not hasattr(self, "_options"):
            self._query = dict(parse_qsl(self.components.query))
        return self._query

    def replace(self, **kwargs: Any) -> "MutableURL":
        if "username" in kwargs or "password" in kwargs or "hostname" in kwargs or "port" in kwargs:
            hostname = kwargs.pop("hostname", self.hostname)
            port = kwargs.pop("port", self.port)
            username = kwargs.pop("username", self.username)
            password = kwargs.pop("password", self.password)

            netloc = hostname
            if port is not None:
                netloc += f":{port}"
            if username is not None:
                userpass = username
                if password is not None:
                    userpass += f":{password}"
                netloc = f"{userpass}@{netloc}"

            kwargs["netloc"] = netloc

        components = self.components._replace(**kwargs)
        return self.__class__(components.geturl())

    def obscure(self) -> str:
        if self.password:
            return self.replace(password="********")._url
        return self._url

    def __str__(self) -> str:
        return self._url

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.obscure())})"

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""

import logging
from functools import partial
from typing import Optional

from django.db import transaction

logger = logging.getLogger(__name__)


def apply_async_on_commit(celery_task, using: Optional[str] = None, *args, **kwargs):
    """
    Apply celery task async and always ignore the task result,
    it will trigger the task on transaction commit when it is in a atomic block.
    """

    fn = partial(celery_task.apply_async, ignore_result=True, *args, **kwargs)

    connection = transaction.get_connection(using)
    if connection.in_atomic_block:
        logger.debug("trigger task %s on transaction commit", celery_task.name)
        transaction.on_commit(fn)

    else:
        logger.debug("trigger task %s immediately", celery_task.name)
        fn()


def delay_on_commit(celery_task, *args, **kwargs):
    """
    Star argument version of `apply_async_on_commit`, does not support the extra options.
    """

    apply_async_on_commit(celery_task, args=args, kwargs=kwargs)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from typing import Optional

from requests import Response


class RequestError(Exception):
    """服务请求异常"""

    def __init__(self, message: str, code: str = "400", response: Optional[Response] = None) -> None:
        super().__init__(message)
        self.message = message
        self.code = code
        self.response = response

    def __str__(self):
        return self.message


class ObjectAlreadyExists(RequestError):
    """该对象已存在."""


class DownloadFailedError(Exception):
    """文件下载失败"""

    def __init__(self, key: str, dest: str):
        """
        :param str key: The name of the key to download from.
        :param str dest: The path to the file to download to.
        """
        super().__init__(key, dest)
        self.key = key
        self.dest = dest


class UploadFailedError(Exception):
    """文件上传失败"""

    def __init__(self, key: str, src: str):
        """
        :param str key: The name of the key to upload to.
        :param str src: The path to the file to upload.
        """
        super().__init__(key, src)
        self.key = key
        self.src = src



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""

import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Type

from celery import shared_task

"""module for creating long-running polling tasks via celery"""
logger = logging.getLogger(__name__)


class PollingStatus(int, Enum):
    """Status of a single polling action"""

    DOING = 1
    DONE = 2


class PollingResult:
    """A single polling result object

    :param status: status of current polling action, such as: DOING, DONE
    :param data: extra data of current polling
    """

    def __init__(self, status: PollingStatus, data: Optional[Any] = None):
        self.status = status
        self.data = data

    def __str__(self):
        return f'stauts={self.status} data={self.data}'

    @classmethod
    def doing(cls, *args, **kwargs):
        """Shortcut for creating doing result"""
        return cls(PollingStatus.DOING, *args, **kwargs)

    @classmethod
    def done(cls, *args, **kwargs):
        """Shortcut for creating done result"""
        return cls(PollingStatus.DONE, *args, **kwargs)


@dataclass
class PollingMetadata:
    """Metadata of a polling process"""

    retries: int
    # unix timestamp of when current polling has been started
    query_started_at: float
    queried_count: int
    # data attribute of last polling action
    last_polling_data: Optional[Dict] = None


class TaskPoller(ABC):
    """task status poller

    :param params: params to perform polling
    :param metadata: metadata object of current poller
    """

    _registered_pollers: Dict[str, Type['TaskPoller']] = {}

    max_retries_on_error = 10
    overall_timeout_seconds = 3600 * 24 * 7
    default_retry_delay_seconds = 10

    def __init__(self, params: Dict, metadata: PollingMetadata):
        self.params = params
        self.metadata = metadata

    def __init_subclass__(cls, *args, **kwargs):
        cls._registered_pollers[cls.__name__] = cls

    @classmethod
    def get_poller_cls(cls, name: str) -> Type['TaskPoller']:
        return cls._registered_pollers[name]

    @classmethod
    def start(cls, params: Dict, callback_handler_cls: Optional[Type] = None):
        """Start a new polling task

        :param params: params for starting polling, must be Json compatible in order to use celery
        :param callback_handler_cls: type to handle poll result
        """
        handler_name = None
        if callback_handler_cls is not None:
            assert issubclass(callback_handler_cls, CallbackHandler)
            handler_name = callback_handler_cls.__name__

        # Start background task
        cls.get_async_task().delay(cls.__name__, handler_name, params)

    def make_next_metadata(self, has_error: bool = False, last_polling_data: Optional[Dict] = None) -> PollingMetadata:
        """Make the metadata object for next polling

        :param has_error: if has error, will affect `retries` attribute
        :param last_polling_data: current polling data, "last" data for next polling
        """
        # Reset retries when no exception occurs
        if not has_error:
            retries = 0
        else:
            retries = self.metadata.retries + 1
        return PollingMetadata(
            retries=retries,
            query_started_at=self.metadata.query_started_at,
            queried_count=self.metadata.queried_count + 1,
            last_polling_data=last_polling_data,
        )

    @abstractmethod
    def query(self) -> PollingResult:
        """Start a polling action, subclasses must override this method"""
        raise NotImplementedError()

    def get_retry_delay(self) -> int:
        """Get delay of next retry"""
        return self.default_retry_delay_seconds

    def exceeded_timeout(self) -> bool:
        """Check if current polling procedure has exceeded max timeout"""
        return (time.time() - self.metadata.query_started_at) > self.get_overall_timeout_seconds()

    def get_overall_timeout_seconds(self) -> int:
        """The overall timeout seconds for a complete polling procedure"""
        return self.overall_timeout_seconds

    def exceeded_max_retries(self) -> bool:
        """Check if current polling has retried too many times"""
        return (self.metadata.retries + 1) > self.max_retries_on_error

    def __str__(self):
        return '<%s: params=%s>' % (self.__class__.__name__, self.params)

    @classmethod
    def get_async_task(self) -> Any:
        """Return the async celery task object for polling in backend"""
        return check_status_until_finished


class CallbackStatus(int, Enum):
    """Status of a finished polling"""

    NORMAL = 0
    TIMEOUT = 1
    EXCEPTION = 2

    def is_exception(self):
        return self in (self.TIMEOUT, self.EXCEPTION)


class CallbackResult:
    """The final result of a polling procedure

    :param status: status of current result
    :param data: data of current result
    :param message: extra message of current result
    """

    def __init__(self, status: CallbackStatus, data: Optional[Dict] = None, message: str = ""):
        self.status = status
        self.data = data or {}
        self.message = message

    @property
    def is_exception(self):
        return self.status.is_exception()

    def to_dict(self):
        return {'status': self.status.value, 'message': self.message, 'data': self.data}

    def __str__(self):
        return '<%s: %s is_exception=%s>' % (self.__class__.__name__, self.to_dict(), self.is_exception)


class CallbackHandler(ABC):
    """handle callback result

    :params: params of current polling result
    """

    _registered_handlers: Dict[str, Type['CallbackHandler']] = {}

    def __init_subclass__(cls, *args, **kwargs):
        cls._registered_handlers[cls.__name__] = cls

    @classmethod
    def get_handler_cls(cls, name: str) -> Type['CallbackHandler']:
        return cls._registered_handlers[name]

    @abstractmethod
    def handle(self, result: CallbackResult, poller: TaskPoller):
        """Handle final callback result

        :param result: CallbackResult instance
        :param poller: Current TaskPoller instance
        """
        raise NotImplementedError()


class NullResultHandler(CallbackHandler):
    """A null implementation of callback result handler"""

    def handle(self, result: CallbackResult, poller: TaskPoller):
        pass


@shared_task(acks_late=True, name='poll_task.check_status_until_finished')
def check_status_until_finished(poller_name: str, handler_name: str, params: Dict, queue: Optional[str] = None):
    """Main async task for polling

    :param poller_name: name of poller class
    :param handler_name: name of result handler
    :param params: params for performing polling
    :param queue: dedicated queue name
    """
    req = check_status_until_finished.request
    metadata = PollingMetadata(
        retries=req.retries,
        query_started_at=req.get('query_started_at', time.time()),
        queried_count=req.get('queried_count', 0),
        last_polling_data=req.get('last_polling_data'),
    )

    # Make handler and poller by name
    poller = TaskPoller.get_poller_cls(poller_name)(params, metadata)
    if handler_name is not None:
        handler_cls = CallbackHandler.get_handler_cls(handler_name)
    else:
        handler_cls = NullResultHandler

    scheduler = PollTaskScheduler(poller, handler_cls)
    next_metadata = scheduler.run()

    if next_metadata:
        # Start next polling
        countdown = poller.get_retry_delay()
        logger.debug('Will retry query status for %s after %s seconds. metadata=%s', poller, countdown, metadata)
        poller.get_async_task().subtask(
            args=(poller_name, handler_name, params),
            kwargs={'queue': queue},
            countdown=countdown,
            retries=next_metadata.retries,
            queue=queue,
        ).apply_async(
            headers={
                'queried_count': next_metadata.queried_count,
                'query_started_at': next_metadata.query_started_at,
                'last_polling_data': next_metadata.last_polling_data,
            }
        )


class PollingQueryError(Exception):
    """Error when perform poller.query method"""


class PollTaskScheduler:
    """Schedule poll tasks"""

    def __init__(self, poller: TaskPoller, handler_cls: Type[CallbackHandler]):
        self.poller = poller
        self.handler_cls = handler_cls

    def run(self) -> Optional[PollingMetadata]:
        """Start schedule process"""
        if self.poller.exceeded_timeout():
            logger.info('exceeded total timeout, ts_query_started=%s' % self.poller.metadata.query_started_at)
            self._callback_timeout()
            return None

        try:
            polling_result = self._safe_query(self.poller)
        except PollingQueryError as e:
            if self.poller.exceeded_max_retries():
                self._callback_exception(e)
                return None

            # Retry next polling, set `last_polling_data` field to the value of last succeeded call
            metadata = self.poller.make_next_metadata(
                has_error=True, last_polling_data=self.poller.metadata.last_polling_data
            )
            return metadata

        if polling_result.status == PollingStatus.DONE:
            ret = CallbackResult(status=CallbackStatus.NORMAL, data=polling_result.data)
            self._callback(ret)
            return None

        metadata = self.poller.make_next_metadata(has_error=False, last_polling_data=polling_result.data)
        return metadata

    @staticmethod
    def _safe_query(poller: TaskPoller) -> PollingResult:
        """call poller's query method with exception handling

        :raises: PollingQueryError
        """
        try:
            polling_result = poller.query()
        except Exception as e:
            logger.exception('Exception when query status, poll_class=%s' % poller)
            raise PollingQueryError(str(e))

        logger.debug('Query status result, poll_class=%s, polling result: %s' % (poller, polling_result))
        return polling_result

    def _callback(self, result: CallbackResult):
        """Callback handler"""
        self.handler_cls().handle(result, self.poller)

    def _callback_timeout(self):
        """Callback handler with timeout result"""
        ret = CallbackResult(status=CallbackStatus.TIMEOUT, message='exceeded total timeout')
        self.handler_cls().handle(ret, self.poller)

    def _callback_exception(self, e: Exception):
        """Callback handler when exceeds max retries"""
        ret = CallbackResult(status=CallbackStatus.EXCEPTION, message=f'exception: {e}')
        self.handler_cls().handle(ret, self.poller)



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
# Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
This command is similar to Django's built-in "loaddata" command with one difference, it expands
environment variables in the fixture data.

WARNING: This command is deprecated and might be removed in the future, please don't use it.
"""

import os
from contextlib import contextmanager

from django.core.management.commands.loaddata import Command as OriginCommand
from django.core.serializers import base
from django.core.serializers import python as _origin_module
from django.core.serializers.python import _get_model
from django.db import DEFAULT_DB_ALIAS, models

SKIP_ENV_VAR_RENDER_TERM = "SKIP_ENV_VAR_RENDER"


def Deserializer(object_list, *, using=DEFAULT_DB_ALIAS, ignorenonexistent=False, **options):  # noqa
    """
    Deserialize simple Python objects back into Django ORM instances.

    It's expected that you pass the Python objects themselves (instead of a
    stream or a string) to the constructor
    """
    handle_forward_references = options.pop("handle_forward_references", False)
    field_names_cache = {}  # Model: <list of field_names>

    for d in object_list:
        # Look up the model and starting build a dict of data for it.
        try:
            Model = _get_model(d["model"])  # noqa: N806
        except base.DeserializationError:
            if ignorenonexistent:
                continue
            else:
                raise
        data = {}
        if "pk" in d:
            try:
                data[Model._meta.pk.attname] = Model._meta.pk.to_python(d.get("pk"))
            except Exception as e:
                raise base.DeserializationError.WithData(e, d["model"], d.get("pk"), None)
        m2m_data = {}
        deferred_fields = {}

        if Model not in field_names_cache:
            field_names_cache[Model] = {f.name for f in Model._meta.get_fields()}
        field_names = field_names_cache[Model]

        # Handle each field
        for field_name, field_value in d["fields"].items():
            if ignorenonexistent and field_name not in field_names:
                # skip fields no longer on model
                continue

            field = Model._meta.get_field(field_name)

            # Handle M2M relations
            if field.remote_field and isinstance(field.remote_field, models.ManyToManyRel):
                try:
                    values = base.deserialize_m2m_values(field, field_value, using, handle_forward_references)
                except base.M2MDeserializationError as e:
                    raise base.DeserializationError.WithData(e.original_exc, d["model"], d.get("pk"), e.pk)
                if values == base.DEFER_FIELD:
                    deferred_fields[field] = field_value
                else:
                    m2m_data[field.name] = values
            # Handle FK fields
            elif field.remote_field and isinstance(field.remote_field, models.ManyToOneRel):
                try:
                    value = base.deserialize_fk_value(field, field_value, using, handle_forward_references)
                except Exception as e:
                    raise base.DeserializationError.WithData(e, d["model"], d.get("pk"), field_value)
                if value == base.DEFER_FIELD:
                    deferred_fields[field] = field_value
                else:
                    data[field.attname] = value
            # Handle all other fields
            else:
                try:
                    if isinstance(field_value, (str, bytes)) and _should_render_by_env_vars():
                        field_value = os.path.expandvars(field_value)  # noqa: PLW2901
                    data[field.name] = field.to_python(field_value)
                except Exception as e:
                    raise base.DeserializationError.WithData(e, d["model"], d.get("pk"), field_value)

        obj = base.build_instance(Model, data, using)
        yield base.DeserializedObject(obj, m2m_data, deferred_fields)


def _should_render_by_env_vars():
    return not os.environ.get(SKIP_ENV_VAR_RENDER_TERM, False)


@contextmanager
def patch(owner, attr, value):
    """Monkey patch context manager.

    with patch(os, 'open', myopen):
        ...
    """
    old = getattr(owner, attr)
    setattr(owner, attr, value)
    try:
        yield getattr(owner, attr)
    finally:
        setattr(owner, attr, old)


class Command(OriginCommand):
    def handle(self, *fixture_labels, **options):
        with patch(_origin_module, "Deserializer", Deserializer):
            super().handle(*fixture_labels, **options)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from typing import List

from blue_krill.monitoring.probe.base import Issue, VirtualProbe
from blue_krill.redis_tools.sentinel import SentinelBackend

try:
    import redis
except ImportError as _e:
    raise ImportError('Error loading redis module: %s.\n' 'Did you install redis?' % _e) from _e


logger = logging.getLogger(__name__)


class RedisProbe(VirtualProbe):
    """
    Usage:
        class SomeRedisProbe(RedisProbe):
            name: str = "some"
            redis_url: str = "redis://localhost:6379/0"

    Example redis url::

        redis://[:password]@localhost:6379/0
        unix://[:password]@/path/to/socket.sock?db=0
    """

    redis_url: str

    def diagnose(self) -> List[Issue]:
        try:
            # from_url already used connection pool
            connection = redis.Redis.from_url(self.redis_url)
            connection.ping()
        except Exception as e:  # pylint: disable=broad-except
            logger.exception("Unknown Exception when connecting to %s", self.redis_url)
            return [Issue(fatal=True, description=str(e))]
        return []


class RedisSentinelProbe(VirtualProbe):
    """
    Usage:
        class SomeRedisSentinelProbe(RedisSentinelProbe):
            name: str = "some"
            redis_url: str = "sentinel://0.0.0.0:26347/0"
            master_name: str = "mycluster"
            sentinel_kwargs: dict = {'password': 'xxxx'}

    Example redis sentinel url::
        sentinel://[:password]@localhost:26347/0
        sentinel://[:password]@0.0.0.0:26347/3;sentinel://[:password]@0.0.0.0:26347/3
    """

    redis_url: str
    master_name: str
    sentinel_kwargs: dict

    def diagnose(self) -> List[Issue]:
        try:
            backend = SentinelBackend(self.redis_url, self.master_name, self.sentinel_kwargs)
            backend.client.ping()
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(
                "Unknown Exception when connecting to %s, sentinel master_name is %s", self.redis_url, self.master_name
            )
            return [Issue(fatal=True, description=str(e))]
        return []



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import datetime
import logging
import random
from contextlib import contextmanager
from dataclasses import dataclass, field
from operator import attrgetter
from threading import RLock
from typing import Any, Generic, Iterable, List, Optional, Tuple, Type, TypeVar

from .exceptions import NoEndpointAvailable
from .ha_algorithm import BasicHAAlgorithm

logger = logging.getLogger(__name__)

T = TypeVar('T')

SUCCESS_SCORE_DELTA = 1
FAILURE_SCORE_DELTA = 50


class HAEndpointPool(Generic[T]):
    """A container type which holds a collection of endpoints, provides HA-related functions,
    e.g. fail(), elect() ect.

    :param items: raw items, each will be bound with an Endpoint object
    :param algorithm: HA Algorithm object, default value: `BasicHAAlgorithm`
    """

    active_endpoint: 'Endpoint'

    def __init__(self, items: Iterable[T], algorithm: 'BasicHAAlgorithm' = BasicHAAlgorithm()):
        self._rlock = RLock()
        self.algorithm = algorithm
        self.endpoints: 'List[Endpoint]' = []

        # Initialize endpoints and start first election
        for item in items:
            self.add(item)
        self.elect()

    def add(self, item: T):
        """Add an endpoint object"""
        self.endpoints.append(Endpoint(raw=item))

    def elect(self, endpoint: 'Optional[Endpoint]' = None):
        """Set `self.active_endpoint` to the best endpoint, it should be the healthy one with
        highest score.

        :param endpoint: Set given Endpoint directly, ignore any algorithms
        """
        with self._rlock:
            # Try to recover unhealthy endpoints
            for ep in self.endpoints:
                if self.algorithm.should_recover(ep):
                    ep.set_healthy()

            if endpoint:
                self.active_endpoint = endpoint
                return

            # Get endpoints with maximum score
            eps = sorted(self.list_healthy(), key=attrgetter('score'), reverse=True)
            if not eps:
                raise NoEndpointAvailable('No healthy endpoint')

            best = random.choice([x for x in eps if x.score == eps[0].score])
            self.active_endpoint = best
            logger.debug(f"Election finished, best: {best}")

    def get(self) -> T:
        """Get current active item"""
        return self.active_endpoint.raw

    def get_endpoint(self) -> 'Endpoint':
        """Get current active endpoint object"""
        return self.active_endpoint

    def list_healthy(self) -> 'List[Endpoint]':
        """List all healthy Endpoints"""
        return [ep for ep in self.endpoints if not ep.is_unhealthy()]

    def fail(self, score: int = FAILURE_SCORE_DELTA):
        """Mark current active endpoint as failed"""
        with self._rlock:
            self.active_endpoint.fail(score_delta=score)
            if self.algorithm.is_unhealthy(self.active_endpoint):
                self.active_endpoint.set_unhealthy()

    def succeed(self, score_delta: int = SUCCESS_SCORE_DELTA):
        """Mark current active endpoint as succeeded"""
        with self._rlock:
            self.active_endpoint.succeed(score_delta=score_delta)

    @contextmanager
    def once(self, failure_excs: Tuple[Type[Exception], ...] = ()) -> Any:
        """A simple context manger which updates endpoint's status automatically. Basically it
        marks current endpoint as "succeed" when everything goes smoothly. But when the code raises
        an exception, it will be captured silently and current endpoint will be marked as "failure"
        if the exception was an instance of `failure_excs`, otherwise it will be raised as is.

        :param failure_excs: Exception types which will be captured.
        """
        # Always call elect() when context begins
        self.elect()
        try:
            yield self.get()
        except Exception as e:
            if isinstance(e, failure_excs):
                logger.warning("Got exception: %s, mark the active endpoint %s as failure", e, self.active_endpoint)
                self.fail()
            else:
                raise
        else:
            self.succeed()

    def __repr__(self) -> str:
        return f'HAEndpointPool(active_endpoint={self.active_endpoint!r}, endpoints={self.endpoints!r})'


@dataclass(order=True)
class Endpoint:
    raw: Any = field(compare=False)
    # Avoid some nodes' scores would be extremely high for the long-term stable,
    # causing error switching time too long.
    max_score: int = 150
    min_score: int = -50

    # score could extend for weight
    # NOTE: do not change _score directly, use fail() & success() please
    _score: int = 100
    success_count: int = 0
    failure_count: int = 0

    # When current endpoint was marked as "unhealthy"(not available)
    unhealthy_at: Optional[datetime.datetime] = None

    def __repr__(self):
        return repr(self.raw) + f"-score<{self._score}>"

    def __hash__(self):
        return hash(repr(self))

    @property
    def score(self):
        return self._score

    def fail(self, score_delta: int = 0):
        """
        :param score_delta: delta of score
        :return: no return
        """
        self._update_score(score_delta=score_delta)
        self.failure_count += 1

    def succeed(self, score_delta: int = 0):
        self._update_score(score_delta=score_delta, fail=False)
        self.success_count += 1

    def is_unhealthy(self):
        """Whether current endpoint is unhealthy"""
        return self.unhealthy_at is not None

    def set_unhealthy(self):
        """Set current endpoint as unhealthy"""
        self.unhealthy_at = datetime.datetime.now()

    def set_healthy(self):
        """Set current endpoint as healthy"""
        self.failure_count = 0
        self.unhealthy_at = None

    def _adjust_max_min(self):
        if self._score >= self.max_score:
            self._score = self.max_score
            return

        if self._score <= self.min_score:
            self._score = self.min_score
            return

    def _update_score(self, score_delta: int, fail: bool = True):
        if score_delta < 0:
            raise ValueError("score delta should not be negative")

        if fail:
            self._score -= score_delta
        else:
            self._score += score_delta

        # make sure score will not beyond preset interval
        self._adjust_max_min()



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import abc
import fnmatch
import json
import logging
import os
import shutil
import stat
import sys
import time
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from textwrap import dedent
from typing import Any, Collection, Dict, Iterator, List, Optional, Set, Type

import click
import toml
from pydantic import BaseModel, Field, ValidationError
from toml.decoder import TomlDecodeError
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

# Set up logging as a console output
hdr = logging.StreamHandler()
hdr.setFormatter(logging.Formatter("[%(asctime)s] %(levelname)s: %(message)s"))
logger = logging.getLogger(__name__)
logger.addHandler(hdr)

__version__ = "0.1.0"

# The default file name of config file
DEFAULT_CONFIG_FILE_NAME = "editionctl.toml"


class FileLinker(abc.ABC):
    @abc.abstractmethod
    def link(self, src_file: PathLike, dst_file: PathLike):
        pass

    @abc.abstractmethod
    def unlink(self, dst_file: PathLike):
        pass


class CopyLinker(FileLinker):
    """Directly copy"""

    def link(self, src_file, dst_file):
        shutil.copyfile(src_file, dst_file, follow_symlinks=False)

    def unlink(self, dst_file):
        Path(dst_file).unlink()


class ReadOnlyCopyLinker(FileLinker):
    """CopyLinker, make sure dst-file is read only"""

    read_only_mode = stat.S_IREAD | stat.S_IRGRP | stat.S_IROTH
    writable_mode = stat.S_IWRITE | stat.S_IREAD | stat.S_IRGRP | stat.S_IROTH

    def set_file_permission(self, path: Path, mode):
        file = Path(path)
        if file.exists():
            file.chmod(mode)

    def link(self, src_file, dst_file):
        # make sure file is writable before copy
        self.set_file_permission(dst_file, self.writable_mode)
        shutil.copyfile(src_file, dst_file, follow_symlinks=False)
        # make sure file is read-only to avoid unexpected modifies
        self.set_file_permission(dst_file, self.read_only_mode)

    def unlink(self, dst_file):
        # Avoid windows cannot delete read-only files
        self.set_file_permission(dst_file, self.writable_mode)
        Path(dst_file).unlink()


class SymbolLinker(FileLinker):
    """Create symbol link from src_file to dst_file"""

    def link(self, src_file, dst_file):
        src = Path(src_file).absolute()
        dst = Path(dst_file).absolute()
        if dst.exists():
            if src.samefile(dst):
                return

            dst.unlink()

        dst.symlink_to(src)

    def unlink(self, dst_file):
        Path(dst_file).unlink()


def get_linker(type_: str) -> FileLinker:
    type_map: Dict[str, Type[FileLinker]] = {
        "default": ReadOnlyCopyLinker,
        "copy": CopyLinker,
        "symbol-link": SymbolLinker,
    }
    return type_map[type_]()


@click.group()
@click.option("--log-level", default="INFO", type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]))
@click.option(
    "--settings-path",
    required=False,
    help='Path fo config file, by default editionctl will try to use "editionctl.toml" in current directory',
)
@click.pass_context
def main(ctx, settings_path, log_level):
    logger.setLevel(getattr(logging, log_level))

    # Read and set command configuration
    if not settings_path:
        settings_path = Path.cwd() / DEFAULT_CONFIG_FILE_NAME

    ctx.ensure_object(dict)
    ctx.obj["settings_path"] = settings_path


class EditionConf(BaseModel):
    """Edition config object

    :param name: name of edition
    :param rel_directory: relative directory path of edition, default to `name`
    """

    name: str
    rel_directory: Optional[str] = None

    def get_rel_directory(self) -> str:
        return self.rel_directory or self.name


class Configuration(BaseModel):
    """The monolith config object for editionctl tool

    :param project_root: if not given, use current directory by default
    :param editions_root: if not given, use `project_root` by default
    :param linker_type: "default" or "symbol-link"
    """

    project_root: Optional[Path] = None
    editions_root: Optional[Path] = None
    linker_type: Optional[str] = None
    editions: List[EditionConf] = Field(default_factory=list)

    def get_project_root(self) -> Path:
        return self.project_root or Path.cwd()

    def get_editions_root(self) -> Path:
        return self.editions_root or self.get_project_root()

    def get_linker_type(self) -> str:
        return self.linker_type or "default"

    def get_edition(self, name: str) -> EditionConf:
        """Get edition conf object by name"""
        for obj in self.editions:
            if obj.name == name:
                return obj
        raise KeyError(f"Edition with name: {name} not found")

    def get_edition_directory(self, name: str) -> Path:
        """Get edition's directory"""
        rel_path = self.get_edition(name).get_rel_directory()
        return self.get_editions_root() / rel_path


def load_configuration(settings_path: PathLike, **overwrites) -> Configuration:
    """Load configuration from a settings file

    :raises: FileNotFoundError, TomlDecodeError, ValidationError(pydantic)
    """
    settings_data = toml.load(settings_path)
    settings_data.update(overwrites)
    return Configuration(**settings_data)


def get_configuration_or_quit(settings_path: PathLike, **overwrites) -> Configuration:
    """Get configuration object or abort execution"""
    try:
        return load_configuration(settings_path, **overwrites)
    except FileNotFoundError:
        logger.critical('Settings file: "%s" not found, exit not.', settings_path)
        sys.exit(1)
    except TomlDecodeError as e:
        logger.critical('Settings file: "%s" is not a valid TOML file, detail: %s', settings_path, e)
        sys.exit(1)
    except ValidationError as e:
        logger.critical('Settings file: "%s" is not a valid TOML file, detail: %s', settings_path, e)
        sys.exit(1)
    except Exception as e:  # pylint: disable=broad-except
        logger.critical("Unable to process settings file, file: %s, error detail: %s", settings_path, e)
        sys.exit(1)


@main.command()
@click.argument("edition_name_in_pos", required=False)
@click.option("--edition-name", required=False)
@click.option("--linker-type", required=False)
@click.pass_context
def activate(ctx, edition_name_in_pos, edition_name, linker_type):
    """Try to active a given edition.

    In the beginning, the plan was to using "namespace packages". But after some experiments,
    we found that solution is too difficult and fragile to implement because we need our namespace
    packages to be nested.

    So currently we are using a simpler(and stupidier) approach: sync all source files under the
    specified edition directory to the main package.
    """
    settings = {}
    # if linker_type has been set, overwrite the settings linker_type
    if linker_type:
        settings["linker_type"] = linker_type

    config = get_configuration_or_quit(ctx.obj["settings_path"], **settings)

    edition_name = edition_name_in_pos or edition_name
    if not edition_name:
        logger.critical('Must provide edtion name via position argument or "--edition-name"')
        sys.exit(1)

    try:
        migrator = EditionFileMigrater(config, edition_name)
        migrator.migrate()
    except RuntimeError:
        logger.exception("unable to finish the migration")
    except ConfigurationError as err:
        logger.critical("Configuration error: %s", err)
    else:
        logger.info(f"Edition {edition_name} activated, linker is {config.get_linker_type()}")


# Exceptions start


class BaseCommandException(Exception):
    pass


class ConfigurationError(BaseCommandException):
    """Error caused by configuration error"""


class MetadataNotFound(BaseCommandException):
    """Can not found a metadata file in given project root"""


class InvalidMetadataFile(BaseCommandException):
    """Can not found a metadata file in given project root"""


# Exceptions end


class EditionMetaData:
    """The metadata for project edition, it was maintained by editionctl automatically."""

    default_gitignore_path = ".gitignore"
    default_metadata_path = "edition-metadata.json"

    def __init__(self, edition_name: str, project_root: PathLike, linker_type: str):
        self.edition_name = edition_name
        self.project_root = Path(project_root)
        self.metadata_path = self.project_root / self.default_metadata_path
        self.gitignore_path = self.project_root / self.default_gitignore_path
        self.external_files: List[Dict[str, Any]] = []
        self.linker_type = linker_type

    @classmethod
    def from_existed_project(cls, project_root: PathLike):
        """Generate a metadata object from an existed project"""
        metadata_path = Path(project_root) / cls.default_metadata_path
        if not metadata_path.exists():
            raise MetadataNotFound()

        try:
            with open(metadata_path, "r") as fp:
                data = json.loads(fp.read())
                result = cls(
                    edition_name=data["edition_name"],
                    project_root=project_root,
                    linker_type=data.get("linker_type", "default"),
                )
                result.external_files = data["external_files"]
                return result
        except Exception:
            logger.exception("Unable to load metadata file")
            raise InvalidMetadataFile("metadata file is invalid")

    def add_files(self, relative_paths: Collection[Path]):
        """Add new files to metadata

        :param relative_path: Relative path of file
        """
        for rel_path in relative_paths:
            filepath = self.project_root / rel_path
            stats = filepath.stat()
            self.external_files.append(
                {"file": str(rel_path), "size": int(stats.st_size), "mtime": int(stats.st_mtime)}
            )

    def save(self):
        """Save the current metadata to project"""
        logger.debug("Writing metadata file...")
        with open(self.metadata_path, "w") as fp:
            content = json.dumps(
                {
                    "edition_name": self.edition_name,
                    "linker_type": self.linker_type,
                    "external_files": self.external_files,
                },
                indent=4,
            )
            fp.write(content)

        # Also write .gitignore file in project_root to avoid push extra edition-specified content
        # by accident
        with open(self.gitignore_path, "w") as fp:
            fp.write(
                "# This file is written by editionctl\n{}".format(
                    "\n".join(obj["file"] for obj in self.external_files)
                )
            )

    def list_managed_files(self) -> Iterator[Path]:
        """Generate all files managed by current metadata"""
        for file_obj in self.external_files:
            yield self.project_root / file_obj["file"]

        yield self.gitignore_path
        yield self.metadata_path


class EditionFileMigrater:
    """The file migrater"""

    def __init__(self, config: Configuration, edition_name: str):
        self.config = config
        self.edition_name = edition_name
        self.linker = get_linker(config.get_linker_type())

    def get_edition_conf(self) -> EditionConf:
        try:
            return self.config.get_edition(self.edition_name)
        except KeyError as e:
            raise RuntimeError("Can not read edition config: %s", e)

    def migrate(self):
        """Migration edition specified files to project root directory

        :raises: ConfigurationError
        """
        # Check if edition directory exists, so we could stop at the very beginning.
        edition_dir = self.config.get_edition_directory(self.edition_name)
        if not edition_dir.exists():
            logger.critical(f"Edition directory: {edition_dir} does not exists, please check your files.")
            raise ConfigurationError("Wrong edition dir")

        last_metadata = load_current_metadata(self.config)
        if last_metadata and self.should_reset(last_metadata):
            reset_project(self.config)

        delete_files = []
        if last_metadata:
            delete_files = [Path(f["file"]) for f in last_metadata.external_files]

        result = DirectorySyncer(self.linker).sync_files(
            source_dir=edition_dir,
            target_dir=self.config.get_project_root(),
            delete_if_not_in_source_files=delete_files,
        )

        metadata = EditionMetaData(self.edition_name, self.config.get_project_root(), self.config.get_linker_type())
        metadata.add_files(result.added_files_relative)
        metadata.save()

    def should_reset(self, last_metadata: EditionMetaData):
        """If the migrator should reset all migrated states"""
        if last_metadata.edition_name != self.edition_name:
            return True
        if last_metadata.linker_type != self.config.get_linker_type():
            return True
        return False


@main.command()
@click.pass_context
def reset(ctx):
    """Reset activated edition"""
    config = get_configuration_or_quit(ctx.obj["settings_path"])
    reset_project(config)


def reset_project(config: Configuration):
    """Reset all migration states.

    - Remove metadata and related files
    - Unlink all migarated files(from metadata file)
    """
    metadata = load_current_metadata(config)
    if metadata is None:
        logger.info("No metadata can be found, resetting aborted")
        return

    logger.info("Resetting project.")
    for filepath in metadata.list_managed_files():
        logger.debug(f"Unlinking file {filepath}.")
        try:
            filepath.unlink()
        except FileNotFoundError:
            pass


def load_current_metadata(config: Configuration) -> Optional[EditionMetaData]:
    """Get the last metadata"""
    try:
        return EditionMetaData.from_existed_project(config.get_project_root())
    except MetadataNotFound:
        return None


@dataclass
class SyncResult:
    """Result of one directory sync process"""

    added_files_relative: Set[Path] = field(default_factory=set)
    added_files: Set[Path] = field(default_factory=set)
    deleted_files: Set[Path] = field(default_factory=set)


class DirectorySyncer:
    """A simple directory syncer"""

    ignore_patterns = ("*.pyc", "*.pyo", "CVS", "tmp", ".git", ".svn", "__pycache__")

    def __init__(self, file_linker: FileLinker):
        self.file_linker = file_linker

    def should_ignore(self, name) -> bool:
        """Should ignore this file/directory or not"""
        return any(fnmatch.fnmatch(name, p) for p in self.ignore_patterns)

    def sync_files(
        self, source_dir: Path, target_dir: Path, delete_if_not_in_source_files: Optional[Collection[Path]] = None
    ) -> SyncResult:
        """Sync files in `source_dir` to `target_dir`, return absolute path of all synced files

        :param source_dir: Source directory
        :param target_dir: Target directory
        :param delete_if_not_in_source_files: When files in this list were not presented in current source directory,
            remove the file if it was in target directory, all path must be relative.
        :returns: All synced file paths
        """
        delete_files = delete_if_not_in_source_files or set()
        delete_files = set(delete_files)

        result = SyncResult()
        for root, dirs, raw_files in os.walk(source_dir, followlinks=False):
            # Ignore dirs and files which should be ignored
            # See: https://stackoverflow.com/questions/19859840/excluding-directories-in-os-walk
            dirs[:] = [d for d in dirs if not self.should_ignore(d)]
            files = [f for f in raw_files if not self.should_ignore(f)]

            rel_path = root[len(str(source_dir)) :].lstrip(os.path.sep)
            dst_path = target_dir / rel_path
            if not dst_path.exists():
                dst_path.mkdir()

            for filename in files:
                src_file = Path(root) / filename
                # Ignore symbolic links
                if os.path.islink(src_file):
                    continue
                dst_file = dst_path / filename

                rel_file = Path(rel_path) / filename
                logger.debug(f"Linking file {rel_file}...")
                self.file_linker.link(src_file, dst_file)
                result.added_files.add(dst_file)
                result.added_files_relative.add(rel_file)

        # Remove files if them were not provided in source directory
        if delete_files:
            for file_path in delete_files:
                if file_path.is_absolute():
                    continue
                if file_path not in result.added_files_relative:
                    logger.debug("Delete file %s.", file_path)
                    try:
                        self.file_linker.unlink(target_dir / file_path)
                    except FileNotFoundError:
                        pass
                    result.deleted_files.add(file_path)
        return result


@main.command()
@click.pass_context
def info(ctx):
    """Print current edition information"""
    config = get_configuration_or_quit(ctx.obj["settings_path"])
    try:
        metadata = EditionMetaData.from_existed_project(config.get_project_root())
    except MetadataNotFound:
        logger.critical("No metadata file found in project root, please activate the edition first.")
        return

    logger.info(f"current edition is {metadata.edition_name}, linker is {metadata.linker_type}")


@main.command()
@click.pass_context
def help(ctx):
    """Print version info"""
    logger.info(f"current version: {__version__}")
    logger.info("Example config file:")
    print(
        dedent(
            """
        # 存放主版本源码文件目录，留空默认为当前目录
        # project_root = 'src/'
        # 存放各版本特殊源码文件的目录，留空默认等于 `project_root`
        # editions_root = 'editions/'
        # 版本文件复制方式，可选值："default"（只读文件拷贝） / "symbol-link"（软连接），默认为 "default"
        # linker_type = 'default'

        # 定义各版本名称与源码相对路径
        [[editions]]
        # 必选：版本名称
        name = "TE"
        # 源码在 `editions_root` 里的相对路径，默认与 `name` 相等
        rel_directory = 'te'

        # 定义多个版本
        [[editions]]
        name = "EE"
        # 源码在 `editions_root` 里的相对路径，默认与 `name` 相等
        rel_directory = 'ee'
    """
        )
    )


@main.command()
@click.pass_context
def develop(ctx):
    """Enter develop mode, auto trigger edition activate procedure after files under current
    edtion directory have been modified.
    """
    config = get_configuration_or_quit(ctx.obj["settings_path"])
    project_root = config.get_project_root()
    try:
        metadata = EditionMetaData.from_existed_project(project_root)
    except MetadataNotFound:
        logger.critical("No metadata file found in project root, please activate the edition first.")
        return

    package_path = config.get_edition_directory(metadata.edition_name)
    event_handler = EditionDevelopEventHandler(config, metadata.edition_name)
    observer = Observer()
    observer.schedule(event_handler, str(package_path), recursive=True)
    logger.info(f"Start watching {package_path} directory for edition {metadata.edition_name}...")
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


class EditionDevelopEventHandler(FileSystemEventHandler):
    """Handler will auto re-activate edition when it's content changes"""

    def __init__(self, config: Configuration, edition_name: str):
        self.config = config
        self.edition_name = edition_name
        super().__init__()

    def on_any_event(self, event):
        logger.debug(f"Event {event} detected, will re-activate project edtion")

        logger.info("Going to re-activate edtion...")
        migrator = EditionFileMigrater(self.config, self.edition_name)
        try:
            migrator.migrate()
        except RuntimeError:
            logger.critical("unable to finish the migration")
        logger.info(f"Edition {self.edition_name} re-activated, linker is {self.config.get_linker_type()}")

        logger.info("Inform project dev server to reload.")


if __name__ == "__main__":
    main()
